#!/usr/bin/env python3
"""
lib/domain.py The domain-scoped utility functions
"""

import re
import sys
import importlib
import inspect
from functools import wraps
from types import ModuleType, FunctionType
from typing import Coroutine, Generator
from typing import Dict, List, Tuple
import yaml

from starlette.exceptions import HTTPException

from halfapi.lib import acl
from halfapi.lib.responses import ORJSONResponse, ODSResponse, XLSXResponse, PlainTextResponse, HTMLResponse
# from halfapi.lib.router import read_router
from halfapi.lib.constants import VERBS

from ..logging import logger

class MissingAclError(Exception):
    """ Exception to use when no acl are specified for a route
    """
    pass

class PathError(Exception):
    """ Exception to use when the path for a route is malformed
    """
    pass

class UnknownPathParameterType(Exception):
    """ Exception to use when the path parameter for a route is not supported
    """
    pass

class UndefinedRoute(Exception):
    """ Exception to use when the route definition cannot be found
    """
    pass

class UndefinedFunction(Exception):
    """ Exception to use when a function definition cannot be found
    """
    pass

class NoDomainsException(Exception):
    """ The exception that is raised when HalfAPI is called without domains
    """
    pass

def route_decorator(fct: FunctionType) -> Coroutine:
    """ Returns an async function that can be mounted on a router
    """
    @wraps(fct)
    @acl.args_check
    async def wrapped(request, *args, **kwargs):
        fct_args_spec = inspect.getfullargspec(fct).args
        fct_args_defaults = inspect.getfullargspec(fct).defaults or []
        fct_args_defaults_dict = dict(list(zip(
            reversed(fct_args_spec),
            reversed(fct_args_defaults)
        )))

        fct_args = request.path_params.copy()

        if 'halfapi' in fct_args_spec:
            fct_args['halfapi'] = {
                'user': request.user if
                    'user' in request else None,
                'config': request.scope.get('config', {}),
                'domain': request.scope.get('domain', 'unknown'),
                'cookies': request.cookies,
                'base_url': request.base_url,
                'url': request.url
            }

        if 'data' in fct_args_spec:
            if 'data' in fct_args_defaults_dict:
                fct_args['data'] = fct_args_defaults_dict['data']
            else:
                fct_args['data'] = {}

            fct_args['data'].update(kwargs.get('data', {}))

        if 'out' in fct_args_spec:
            fct_args['out'] = kwargs.get('out')

        """ If format argument is specified (either by get, post param or function argument)
        """
        if 'ret_type' in fct_args_defaults_dict:
            ret_type = fct_args_defaults_dict['ret_type']
        else:
            ret_type = fct_args.get('data', {}).get('format', 'json')

        logger.debug('Return type {} (defaults: {})'.format(ret_type,
            fct_args_defaults_dict))
        try:
            logger.debug('FCT_ARGS***** %s', fct_args)
            if ret_type == 'json':
                return ORJSONResponse(fct(**fct_args))

            if ret_type == 'ods':
                res = fct(**fct_args)
                assert isinstance(res, list)
                for elt in res:
                    assert isinstance(elt, dict)

                return ODSResponse(res)

            if ret_type == 'xlsx':
                res = fct(**fct_args)
                assert isinstance(res, list)
                for elt in res:
                    assert isinstance(elt, dict)

                return XLSXResponse(res)

            if ret_type in ['html', 'xhtml']:
                res = fct(**fct_args)
                assert isinstance(res, str)

                return HTMLResponse(res)

            if ret_type in 'txt':
                res = fct(**fct_args)
                assert isinstance(res, str)

                return PlainTextResponse(res)


            raise NotImplementedError

        except NotImplementedError as exc:
            raise HTTPException(501) from exc
        except Exception as exc:
            # TODO: Write tests
            logger.error(exc, exc_info=True)
            if not isinstance(exc, HTTPException):
                raise HTTPException(500) from exc
            raise exc

    return wrapped


def get_fct_name(http_verb: str, path: str) -> str:
    """
    Returns the predictable name of the function for a route

    Parameters:
        - http_verb (str): The Route's HTTP method (GET, POST, ...)
        - path (str): The functions path

    Returns:
        str: The *unique* function name for a route and it's verb


    Examples:

        >>> get_fct_name('get', '')
        'get'

        >>> get_fct_name('GET', '')
        'get'

        >>> get_fct_name('POST', 'foo')
        'post_foo'

        >>> get_fct_name('POST', 'bar')
        'post_bar'

        >>> get_fct_name('DEL', 'foo/{boo}')
        'del_foo_BOO'

        >>> get_fct_name('DEL', '{boo:zoo}/far')
        'del_BOO_far'
    """
    if path and path[0] == '/':
        path = path[1:]

    fct_name = [http_verb.lower()]
    for elt in path.split('/'):
        if elt and elt[0] == '{':
            fct_name.append(elt[1:-1].split(':')[0].upper())
        elif elt:
            fct_name.append(elt)

    return '_'.join(fct_name)




import pytest
from starlette.responses import PlainTextResponse
from starlette.testclient import TestClient
from halfapi.half_route import HalfRoute
from halfapi.lib import acl

def test_acl_Check(dummy_app, token_debug_false_builder):
    """
    A request with ?check should always return a 200 status code
    """

    @HalfRoute.acl_decorator(params=[{'acl':acl.public}])
    async def test_route_public(request, **kwargs):
        raise Exception('Should not raise')
        return PlainTextResponse('ok')

    dummy_app.add_route('/test_public', test_route_public)
    test_client = TestClient(dummy_app)

    resp = test_client.request('get', '/test_public?check')
    assert resp.status_code == 200

    @HalfRoute.acl_decorator(params=[{'acl':acl.private}])
    async def test_route_private(request, **kwargs):
        raise Exception('Should not raise')
        return PlainTextResponse('ok')

    dummy_app.add_route('/test_private', test_route_private)
    test_client = TestClient(dummy_app)

    resp = test_client.request('get', '/test_private')
    assert resp.status_code == 401

    resp = test_client.request('get', '/test_private?check')
    assert resp.status_code == 200





#!/usr/bin/env python3
"""
cli/routes.py Defines the "halfapi routes" cli commands
"""
# builtins
import sys
import importlib
from pprint import pprint
import orjson

import click

from .cli import cli

from ..logging import logger

# from ..lib.domain import domain_schema_dict
from ..lib.constants import DOMAIN_SCHEMA, ROUTE_SCHEMA
from ..lib.responses import ORJSONResponse

@click.argument('module', required=True)
@click.option('--export', default=False, is_flag=True)
@click.option('--validate', default=False, is_flag=True)
@click.option('--check', default=False, is_flag=True)
@click.option('--noheader', default=False, is_flag=True)
@click.option('--schema', default=False, is_flag=True)
@cli.command()
def routes(module, export=False, validate=False, check=False, noheader=False, schema=False):
    """
    The "halfapi routes" command
    """
    # try:
    #     mod = importlib.import_module(module)
    # except ImportError as exc:
    #     raise click.BadParameter('Cannot import this module', param=module) from exc

    # if export:
    #     click.echo(schema_to_csv(module, header=not noheader))

    # if schema:
    #     routes_d = domain_schema_dict(mod)
    #     ROUTE_SCHEMA.validate(routes_d)
    #     click.echo(orjson.dumps(routes_d,
    #         option=orjson.OPT_NON_STR_KEYS,
    #         default=ORJSONResponse.default_cast))


    # if validate:
    #     routes_d = domain_schema_dict(mod)
    #     try:
    #         ROUTE_SCHEMA.validate(routes_d)
    #     except Exception as exc:
    #         raise exc




import logging


def config_logging(level=logging.INFO):

    # When run by 'uvicorn ...', a root handler is already
    # configured and the basicConfig below does nothing.
    # To get the desired formatting:
    logging.getLogger().handlers.clear()

    # 'uvicorn --log-config' is broken so we configure in the app.
    #   https://github.com/encode/uvicorn/issues/511
    logging.basicConfig(
        # match gunicorn format
        format='%(asctime)s [%(process)d] [%(levelname)s] %(message)s',
        datefmt='[%Y-%m-%d %H:%M:%S %z]',
        level=level)

        # When run by 'gunicorn -k uvicorn.workers.UvicornWorker ...',
    # These loggers are already configured and propogating.
    # So we have double logging with a root logger.
    # (And setting propagate = False hurts the other usage.)
    logging.getLogger('uvicorn.asgi').handlers.clear()
    logging.getLogger('uvicorn.access').handlers.clear()
    logging.getLogger('uvicorn.error').handlers.clear()
    logging.getLogger('uvicorn.asgi').propagate = True
    logging.getLogger('uvicorn.access').propagate = True
    logging.getLogger('uvicorn.error').propagate = True

config_logging()
logger = logging.getLogger()



from .halfapi import HalfAPI
from .conf import CONFIG, SCHEMA
from .logging import logger

logger.info('CONFIG: %s', CONFIG)
logger.info('SCHEMA: %s', SCHEMA)

application = HalfAPI(
    CONFIG, SCHEMA or None).application



# #!/usr/bin/env python3
# import importlib
# from halfapi.lib.domain import VERBS, gen_routes, gen_router_routes, \
#     MissingAclError, domain_schema_dict, domain_schema_list
# 
# from types import FunctionType
# 
# 
# def test_gen_router_routes():
#     from .dummy_domain import routers
#     for path, verb, m_router, fct, params in gen_router_routes(routers, ['dummy_domain']):
#         assert isinstance(path, str)
#         assert verb in VERBS
#         assert len(params) > 0
#         assert hasattr(fct, '__call__')
#         assert len(m_router.__file__) > 0
# 
# 
# def test_gen_routes():
#     from .dummy_domain.routers.abc.alphabet import TEST_uuid
#     try:
#         gen_routes(
#             TEST_uuid,
#             'get',
#             ['abc', 'alphabet', 'TEST_uuid', ''],
#             [])
#     except MissingAclError:
#         assert True
# 
#     fct, params = gen_routes(
#         TEST_uuid,
#         'get',
#         ['abc', 'alphabet', 'TEST_uuid', ''],
#         TEST_uuid.ACLS['GET'])
# 
#     assert isinstance(fct, FunctionType)
#     assert isinstance(params, list)
#     assert len(TEST_uuid.ACLS['GET']) == len(params)
# 
# def test_domain_schema_dict():
#     from .dummy_domain import routers
#     d_res = domain_schema_dict(routers)
# 
#     assert isinstance(d_res, dict)
# 
# def test_domain_schema_list():
#     from .dummy_domain import routers
#     res = domain_schema_list(routers)
# 
#     assert isinstance(res, list)
#     assert len(res) > 0

from starlette.testclient import TestClient
from starlette.responses import Response
from starlette.routing import Router, Route

from halfapi.lib.domain import route_decorator
from halfapi.lib.user import Nobody

def test_route_decorator():
    """ It should decorate an async function that fullfills its arguments
    """
    def route(halfapi, data, out, ret_type='txt'):
        for key in ['user', 'config', 'domain', 'cookies', 'base_url', 'url']:
            assert key in halfapi

        assert halfapi['user'] is None
        assert isinstance(halfapi['config'], dict)
        assert len(halfapi['config']) == 0
        assert isinstance(halfapi['domain'], str)
        assert halfapi['domain'] == 'unknown'
        assert isinstance(halfapi['cookies'], dict)
        assert len(halfapi['cookies']) == 0
        assert len(str(halfapi['base_url'])) > 0
        assert str(halfapi['base_url']) == 'http://testserver/'
        assert len(str(halfapi['url'])) > 0
        assert str(halfapi['url']) == 'http://testserver/'
        assert isinstance(data, dict)
        assert len(data) == 0

        assert out is None

        assert ret_type is 'txt'

        return ''

    async_route = route_decorator(route)
    app = Router([Route('/', endpoint=async_route, methods=['GET'])])
    client = TestClient(app)
    response = client.get('/')
    assert response.is_success
    assert response.content.decode() == ''

    def route(data, out, ret_type='txt'):
        assert isinstance(data, dict)
        assert len(data) == 0

        assert out is None

        assert ret_type is 'txt'

        return ''

    async_route = route_decorator(route)
    app = Router([Route('/', endpoint=async_route, methods=['GET'])])
    client = TestClient(app)
    response = client.get('/')
    assert response.is_success
    assert response.content.decode() == ''

    def route(data):
        assert isinstance(data, dict)
        assert len(data) == 2
        assert data['toto'] == 'tata'
        assert data['bouboul'] == True

        return ''

    async_route = route_decorator(route)
    app = Router([Route('/', endpoint=async_route, methods=['POST'])])
    client = TestClient(app)
    response = client.post('/', json={'toto': 'tata', 'bouboul': True})
    assert response.is_success
    assert response.json() == ''



#!/usr/bin/env python3
# builtins
""" Response module

Contains some base response classes

Classes :
    - HJSONResponse
    - InternalServerErrorResponse
    - NotFoundResponse
    - NotImplementedResponse
    - ORJSONResponse
    - PlainTextResponse
    - ServiceUnavailableResponse
    - UnauthorizedResponse
    - ODSResponse

"""
from datetime import date
import decimal
import typing
from io import BytesIO
import orjson

# asgi framework
from starlette.responses import PlainTextResponse, Response, JSONResponse, HTMLResponse
from starlette.requests import Request
from starlette.exceptions import HTTPException

from .user import JWTUser, Nobody
from ..logging import logger


__all__ = [
    'HJSONResponse',
    'InternalServerErrorResponse',
    'NotFoundResponse',
    'NotImplementedResponse',
    'ORJSONResponse',
    'PlainTextResponse',
    'ServiceUnavailableResponse',
    'UnauthorizedResponse']


class InternalServerErrorResponse(Response):
    """ The 500 Internal Server Error default Response
    """
    def __init__(self, *args, **kwargs):
        super().__init__(status_code=500)


class NotFoundResponse(Response):
    """ The 404 Not Found default Response
    """
    def __init__(self, *args, **kwargs):
        super().__init__(status_code=404)


class NotImplementedResponse(Response):
    """ The 501 Not Implemented default Response
    """
    def __init__(self, *args, **kwargs):
        super().__init__(status_code=501)

class ServiceUnavailableResponse(Response):
    """ The 503 Service Unavailable default Response
    """
    def __init__(self, *args, **kwargs):
        super().__init__(status_code=503)

class UnauthorizedResponse(Response):
    """ The 401 Not Found default Response
    """
    def __init__(self, *args, **kwargs):
        super().__init__(status_code = 401)


class ORJSONResponse(JSONResponse):
    """ The response that encodes data into JSON
    """
    def __init__(self, content, default=None, **kwargs):
        self.default = default if default is not None else ORJSONResponse.default_cast
        super().__init__(content, **kwargs)

    def render(self, content: typing.Any) -> bytes:
        return orjson.dumps(content,
            option=orjson.OPT_NON_STR_KEYS,
            default=self.default)

    @staticmethod
    def default_cast(typ):
        """ Cast the data in JSON-serializable type
        """
        str_types = {
            decimal.Decimal
        }
        list_types = {
            set
        }
        jsonable_types = {
            JWTUser, Nobody
        }

        if callable(typ):
            return typ.__name__
        if type(typ) in str_types:
            return str(typ)
        if type(typ) in list_types:
            return list(typ)
        if type(typ) in jsonable_types:
            return typ.json

        raise TypeError(f'Type {type(typ)} is not handled by ORJSONResponse')


class HJSONResponse(ORJSONResponse):
    """ The response that encodes generator data into JSON
    """
    def render(self, content: typing.Generator):
        return super().render(list(content))

class ODSResponse(Response):
    file_type = 'ods'

    def __init__(self, d_rows: typing.List[typing.Dict]):
        try:
            import pyexcel as pe
        except ImportError:
            """ ODSResponse is not handled
            """
            super().__init__(content=
                'pyexcel is not installed, ods format not available'
            )
            return

        with BytesIO() as ods_file:
            rows = []
            if len(d_rows):
                rows_names = list(d_rows[0].keys())
                for elt in d_rows:
                    rows.append(list(elt.values()))

                rows.insert(0, rows_names)

            self.sheet = pe.Sheet(rows)
            self.sheet.save_to_memory(
                file_type=self.file_type,
                stream=ods_file)

            filename = f'{date.today()}.{self.file_type}'

            super().__init__(
                content=ods_file.getvalue(),
                headers={
                    'Content-Type': 'application/vnd.oasis.opendocument.spreadsheet; charset=UTF-8',
                    'Content-Disposition': f'attachment; filename="{filename}"'},
                status_code = 200)


class XLSXResponse(ODSResponse):
    file_type = 'xlsx'

def gen_exception_route(response_cls):
    async def exception_route(req: Request, exc: HTTPException):
        return response_cls()

    return exception_route



""" Schemas module

Functions :
    - schema_json
    - schema_dict_dom
    - get_acls

Constant :
    SCHEMAS (starlette.schemas.SchemaGenerator)
"""

import os
import importlib
from typing import Dict, Coroutine, List
from types import ModuleType

from starlette.schemas import SchemaGenerator

from .. import __version__
from ..logging import logger
from .routes import api_routes
from .responses import ORJSONResponse

SCHEMAS = SchemaGenerator(
    {"openapi": "3.0.0", "info": {"title": "HalfAPI", "version": __version__}}
)

async def schema_json(request, *args, **kwargs):
    """
    description: Returns the current API routes description (OpenAPI v3)
                 as a JSON object
    """
    return ORJSONResponse(
        SCHEMAS.get_schema(routes=request.app.routes))


def schema_csv_dict(csv: List[str], prefix='/') -> Dict:
    package = None
    schema_d = {}

    modules_d = {}

    acl_modules_d = {}


    for line in csv:
        if not line:
            continue

        path, verb, router, acl_fct_name, args_req, args_opt, out = line.strip().split(';')
        logger.info('schema_csv_dict %s %s %s', path, args_req, args_opt)
        path = f'{prefix}{path}'

        if path not in schema_d:
            schema_d[path] = {}


        if verb not in schema_d[path]:
            mod_str = router.split(':')[0]
            fct_str = router.split(':')[1]

            if mod_str not in modules_d:
                modules_d[mod_str] = importlib.import_module(mod_str)

            if not hasattr(modules_d[mod_str], fct_str):
                raise Exception(
                    'Missing function in module. module:{} function:{}'.format(
                        router, fct_str
                    )
                )

            fct = getattr(modules_d[mod_str], fct_str)

            schema_d[path][verb] = {
                'module': modules_d[mod_str],
                'fct': fct,
                'acls': []
            }

        if package and router.split('.')[0] != package:
            raise Exception('Multi-domain is not allowed in that mode')

        package = router.split('.')[0]
        if not len(package):
            raise Exception(
                'Empty package name (router=%s)'.format(router))

        acl_package = f'{package}.acl'

        if acl_package not in acl_modules_d:
            if acl_package not in modules_d:
                modules_d[acl_package] = importlib.import_module(acl_package)
            if not hasattr(modules_d[acl_package], acl_fct_name):
                raise Exception(
                    'Missing acl function in module. module:{} acl:{}'.format(
                        acl_package, acl_fct_name
                    )
                )

            acl_modules_d[acl_package] = {}

        acl_modules_d[acl_package][acl_fct_name] = getattr(modules_d[acl_package], acl_fct_name)

        schema_d[path][verb]['acls'].append({
            'acl': acl_modules_d[acl_package][acl_fct_name],
            'args': {
                'required': set(args_req.split(',')) if len(args_req) else set(),
                'optional': set(args_opt.split(',')) if len(args_opt) else set()
            }
        })

    return schema_d



import os
import sys
import subprocess
from types import ModuleType
from typing import Dict
from pprint import pprint

from schema import SchemaError
from .constants import VERBS, ROUTER_SCHEMA
from ..logging import logger





# import os
# import pytest
# from schema import SchemaError
# from halfapi.lib.router import read_router
# from halfapi.lib.constants import ROUTER_SCHEMA, ROUTER_ACLS_SCHEMA
# 
# def test_read_router_routers():
#     from .dummy_domain import routers
# 
#     router_d = read_router(routers)
#     assert '' in router_d
#     assert 'SUBROUTES' in router_d['']
#     assert isinstance(router_d['']['SUBROUTES'], list)
# 
#     for elt in os.scandir(routers.__path__[0]):
#         if elt.is_dir():
#             assert elt.name in router_d['']['SUBROUTES']
# 
# def test_read_router_abc():
#     from .dummy_domain.routers import abc
#     router_d = read_router(abc)
# 
#     assert '' in router_d
#     assert 'SUBROUTES' in router_d['']
#     assert isinstance(router_d['']['SUBROUTES'], list)
# 
# def test_read_router_alphabet():
#     from .dummy_domain.routers.abc import alphabet
#     router_d = read_router(alphabet)
# 
#     assert '' in router_d
#     assert 'SUBROUTES' in router_d['']
#     assert isinstance(router_d['']['SUBROUTES'], list)
# 
#     ROUTER_SCHEMA.validate(router_d)
# 
#     with pytest.raises(SchemaError):
#         """ Test that we cannot specify wrong method in ROUTES or ACLS
# 
#         TODO: Write more errors
#         """
#         router_d['']['TEG'] = {}
#         ROUTER_SCHEMA.validate(router_d)
# 
# def test_read_router_TEST():
#     from .dummy_domain.routers.abc.alphabet import TEST_uuid
#     router_d = read_router(TEST_uuid)
# 
#     print(router_d)
#     assert '' in router_d
#     assert 'SUBROUTES' in router_d['']
#     assert isinstance(router_d['']['GET'], list)
#     assert isinstance(router_d['']['POST'], list)
#     assert isinstance(router_d['']['PATCH'], list)
#     assert isinstance(router_d['']['PUT'], list)
# 
# 



#!/usr/bin/env python3
"""
cli/cli.py Main entry point of halfapi cli tool

The init command is the only command loaded if not in a *project dir*, and it is
not loaded otherwise.
"""
# builtins
import click

@click.group(invoke_without_command=True)
@click.option('--version', is_flag=True)
@click.pass_context
def cli(ctx, version):
    """
    HalfAPI Cli entry point

    It uses the Click library
    """
    if version:
        from halfapi import version
        click.echo(version())

from . import config
from . import domain
from . import run
from . import init
from . import routes



"""
Timing module

Helpers to gathers stats on requests timing

class HTimingClient
"""
import logging

from timing_asgi import TimingClient

from ..logging import logger

class HTimingClient(TimingClient):
    """ Used to redefine TimingClient.timing
    """
    def timing(self, metric_name, timing, tags):
        tags_d = dict(map(lambda elt: elt.split(':'), tags))

        logger.debug('[TIME:%s][%s] %s %s - %sms',
            tags_d['time'], metric_name,
            tags_d['http_method'], tags_d['http_status'],
            round(timing*1000, 2))



from uuid import UUID
from starlette.authentication import BaseUser, UnauthenticatedUser

class Nobody(UnauthenticatedUser):
    """ Nobody class

    The default class when no token is passed
    """
    @property
    def json(self):
       return {
        'id' : '',
        'token': '',
        'payload': ''
    }


class JWTUser(BaseUser):
    """ JWTUser class

    Is used to store authentication informations
    """
    def __init__(self, user_id: UUID, token: str, payload: dict) -> None:
        self.__id = user_id
        self.token = token
        self.payload = payload

    def __str__(self):
        return str(self.json)

    @property
    def json(self):
        return {
            'id' : str(self.__id),
            'token': self.token,
            'payload': self.payload
        }

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return ' '.join(
            (self.payload.get('name'), self.payload.get('firstname')))

    @property
    def id(self) -> str:
        return self.__id


class CheckUser(BaseUser):
    """ CheckUser class

    Is used to call checks with give user_id, to know if it passes the ACLs for
    the given route.

    It should never be able to run a route function.
    """
    def __init__(self, user_id: UUID) -> None:
        self.__id = user_id


    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return 'check_user'

    @property
    def id(self) -> str:
        return self.__id







#!/usr/bin/env python3
"""
Base ACL module that contains generic functions for domains ACL
"""
from dataclasses import dataclass
from functools import wraps
from json import JSONDecodeError
from starlette.authentication import UnauthenticatedUser
from starlette.exceptions import HTTPException

from ..logging import logger

def public(*args, **kwargs) -> bool:
    "Unlimited access"
    return True

def private(*args, **kwargs) -> bool:
    "Forbidden access"
    return False


def connected(fct=public):
    """ Decorator that checks if the user object of the request has been set
    """
    @wraps(fct)
    def caller(req, *args, **kwargs):
        if (not hasattr(req, 'user')
          or isinstance(req.user, UnauthenticatedUser)
          or not hasattr(req.user, 'is_authenticated')):
            return False

        if hasattr(req, 'path_params'):
            return fct(req, **{**kwargs, **req.path_params})
        return fct(req, **{**kwargs})

    return caller

def args_check(fct):
    """ Decorator that puts required and optional arguments in scope

    For GET requests it uses the query_params

    For POST requests it uses the body as JSON

    If "check" is present in the query params, nothing is done.

    If some required arguments are missing, a 400 status code is sent.
    """
    @wraps(fct)
    async def caller(req, *args, **kwargs):
        if 'check' in req.query_params:
            # Check query param should not read the "args"
            return await fct(req, *args, **kwargs)

        data_ = {}
        if req.method == 'GET':
            data_ = dict(req.query_params)

        elif req.method in ['POST', 'PATCH', 'PUT', 'DELETE']:
            if req.scope.get('headers'):
                if b'content-type' not in dict(req.scope.get('headers')):
                    data_ = {}
                else:
                    content_type = dict(req.scope.get('headers')).get(b'content-type').decode().split(';')[0]

                    if content_type == 'application/json':
                        try:
                            data_ = await req.json()
                        except JSONDecodeError as exc:
                            logger.debug('Posted data was not JSON')
                            pass
                    elif content_type in [
                        'multipart/form-data', 'application/x-www-form-urlencoded']:
                        data_ = dict(await req.form())
                    else:
                        data_ = await req.body()

        def plural(array: list) -> str:
            return 's' if len(array) > 1 else ''
        def comma_list(array: list) -> str:
            return ', '.join(array)


        args_d = req.scope.get('args')
        if args_d is not None and isinstance(data_, dict):
            required = args_d.get('required', set())

            missing = []
            data = {}

            for key in required:
                data[key] = data_.get(key, None)
                if data[key] is None:
                    missing.append(key)

            if missing:
                raise HTTPException(
                    400,
                    f"Missing value{plural(missing)} for: {comma_list(missing)}!")

            optional = args_d.get('optional', set())
            for key in optional:
                    if key in data_:
                        data[key] = data_[key]
        else:
            """ Unsafe mode, without specified arguments, or plain text mode
            """
            data = data_

        kwargs['data'] = data

        out_s = req.scope.get('out')
        if out_s:
            kwargs['out'] = list(out_s)

        return await fct(req, *args, **kwargs)

    return caller

# ACLS list for doc and priorities
# Write your own constant in your domain or import this one
# Format : (acl_name: str, acl_documentation: str, priority: int, [public=False])
#
# The 'priority' integer is greater than zero and the lower values means more
# priority. For a route, the order of declaration of the ACLs should respect
# their priority.
#
# When the 'public' boolean value is True, a route protected by this ACL is
# defined on the "/halfapi/acls/acl_name", that returns an empty response and
# the status code 200 or 401.
 
ACLS = (
    ('private', private.__doc__, 0, True),
    ('public', public.__doc__, 999, True)
)

@dataclass
class ACL():
    name: str
    documentation: str
    priority: int
    public: bool = False



#!/usr/bin/env python3
"""
conf.py reads the current configuration files

It uses the following environment variables :

    - HALFAPI_CONF_DIR (default: /etc/half_api)


It defines the following globals :

    - PROJECT_NAME (str) - HALFAPI_PROJECT_NAME
    - PRODUCTION (bool) - HALFAPI_PRODUCTION
    - LOGLEVEL (str) - HALFAPI_LOGLEVEL
    - BASE_DIR (str) - HALFAPI_BASE_DIR
    - HOST (str) - HALFAPI_HOST
    - PORT (int) - HALFAPI_PORT
    - CONF_DIR (str) - HALFAPI_CONF_DIR
    - DRYRUN (bool) - HALFAPI_DRYRUN

It reads the following ressource :

    - ./.halfapi/config

It follows the following format :

    [project]
    halfapi_version = HALFAPI_VERSION

    [domain.domain_name]
    name = domain_name
    routers = routers

    [domain.domain_name.config]
    option = Argh

"""

import logging
import os
from os import environ
import sys
import importlib
import tempfile
import uuid

import toml

from .logging import logger

PRODUCTION = True
LOGLEVEL = 'info'
CONF_FILE = os.environ.get('HALFAPI_CONF_FILE', '.halfapi/config')
DRYRUN = bool(os.environ.get('HALFAPI_DRYRUN', False))

SCHEMA = {}

CONF_DIR = environ.get('HALFAPI_CONF_DIR', '/etc/half_api')
HALFAPI_ETC_FILE=os.path.join(
    CONF_DIR, 'config'
)
HALFAPI_DOT_FILE=os.path.join(
    os.getcwd(), '.halfapi', 'config')

HALFAPI_CONFIG_FILES = []

try:
    with open(HALFAPI_ETC_FILE, 'r'):
        HALFAPI_CONFIG_FILES.append(HALFAPI_ETC_FILE)
except FileNotFoundError:
    logger.error('Cannot find a configuration file under %s', HALFAPI_DOT_FILE)

try:
    with open(HALFAPI_DOT_FILE, 'r'):
        HALFAPI_CONFIG_FILES.append(HALFAPI_DOT_FILE)
except FileNotFoundError:
    logger.error('Cannot find a configuration file under %s', HALFAPI_DOT_FILE)


def read_config():
    """
    The highest index in "filenames" are the highest priorty
    """
    d_res = {}

    logger.info('Reading config files %s', HALFAPI_CONFIG_FILES)
    for CONF_FILE in HALFAPI_CONFIG_FILES:
        d_res.update( toml.load(HALFAPI_CONFIG_FILES) )

    logger.info('Reading config files (result) %s', d_res)
    return { **d_res.get('project', {}), 'domain': d_res.get('domain', {}) }

CONFIG = read_config()

PROJECT_NAME = CONFIG.get('project_name',
    environ.get('HALFAPI_PROJECT_NAME', os.getcwd().split('/')[-1]))

if environ.get('HALFAPI_DOMAIN_NAME'):
    DOMAIN_NAME = environ.get('HALFAPI_DOMAIN_NAME')
    if 'domain' in CONFIG and DOMAIN_NAME in CONFIG['domain'] \
        and 'config' in CONFIG['domain'][DOMAIN_NAME]:

        domain_config = CONFIG['domain'][DOMAIN_NAME]['config']
    else:
        domain_config = {}

    CONFIG['domain'] = {}

    CONFIG['domain'][DOMAIN_NAME] = {
        'enabled': True,
        'name': DOMAIN_NAME,
        'prefix': False
    }

    CONFIG['domain'][DOMAIN_NAME]['config'] = domain_config

    if environ.get('HALFAPI_DOMAIN_MODULE'):
        dom_module = environ.get('HALFAPI_DOMAIN_MODULE')
        CONFIG['domain'][DOMAIN_NAME]['module'] = dom_module

if len(CONFIG.get('domain', {}).keys()) == 0:
    logger.info('No domains')

# Bind
HOST = CONFIG.get('host',
    environ.get('HALFAPI_HOST', '127.0.0.1'))
PORT = int(CONFIG.get(
    'port',
    environ.get('HALFAPI_PORT', '3000')))


# Secret
SECRET = CONFIG.get(
    'secret',
    environ.get('HALFAPI_SECRET'))

if not SECRET:
    # TODO: Create a temporary secret
    _, SECRET = tempfile.mkstemp()
    with open(SECRET, 'w') as secret_file:
        secret_file.write(str(uuid.uuid4()))

try:
    with open(SECRET, 'r') as secret_file:
        CONFIG['secret'] = SECRET.strip()
except FileNotFoundError as exc:
    logger.info('Running without secret file: %s', SECRET or 'no file specified')

PRODUCTION = bool(CONFIG.get(
    'production',
    environ.get('HALFAPI_PROD', True)))

LOGLEVEL = CONFIG.get(
    'loglevel',
    environ.get('HALFAPI_LOGLEVEL', 'info')).lower()

BASE_DIR = CONFIG.get(
    'base_dir',
    environ.get('HALFAPI_BASE_DIR', '.'))

CONFIG['project_name'] = PROJECT_NAME
CONFIG['production'] = PRODUCTION
CONFIG['secret'] = SECRET
CONFIG['host'] = HOST
CONFIG['port'] = PORT
CONFIG['dryrun'] = DRYRUN



import os
import sys
import importlib
import subprocess
import time
import pytest
import json
from pprint import pprint
from starlette.routing import Route
from starlette.testclient import TestClient

def test_get_config_route(dummy_project, application_domain):
    c = TestClient(application_domain)
    r = c.request('get', '/')
    assert r.status_code == 200
    pprint(r.json())
    r = c.request('get', '/config')
    assert r.status_code == 200
    pprint(r.json())
    assert 'test' in r.json()

def test_get_route(dummy_project, application_domain):
    c = TestClient(application_domain)
    path = verb = params = None
    dummy_domain_routes = [
        ('config','GET'),
        ('config','GET'),
        ('async_router/abc/pinnochio','GET'),
        ('async_router/config','GET'),
        # ('abc/pinnochio','GET'),
        # ('abc/alphabet','GET'),
    ]

    for route_def in dummy_domain_routes:
        path, verb = route_def[0], route_def[1]
        route_path = '/{}'.format(path)
        print(route_path)
        try:
            if verb.lower() == 'get':
                r = c.request('get', route_path)
            elif verb.lower() == 'post':
                r = c.request('post', route_path)
            elif verb.lower() == 'patch':
                r = c.request('patch', route_path)
            elif verb.lower() == 'put':
                r = c.request('put', route_path)
            elif verb.lower() == 'delete':
                r = c.request('delete', route_path)
            else:
                raise Exception(verb)
            try:
                assert r.status_code in [200, 501]
            except AssertionError as exc:
                print('{} [{}] {}'.format(str(r.status_code), verb, route_path))
                raise exc from exc

        except NotImplementedError:
            pass

    dummy_domain_path_routes = [
        ('abc/alphabet/{test}','GET'),
    ]

    #for route_def in dummy_domain_path_routes:
    for route_def in []:#dummy_domain_routes:
        from uuid import uuid4
        test_uuid = uuid4()
        for route_def in dummy_domain_path_routes:
            path, verb = route_def[0], route_def[1]
            path = path.format(test=str(test_uuid))
            route_path = f'/{path}'
            if verb.lower() == 'get':
                r = c.request('get', f'{route_path}')

            assert r.status_code == 200


def test_delete_route(dummy_project, application_domain):
    c = TestClient(application_domain)
    from uuid import uuid4
    arg = str(uuid4())
    r = c.request('delete', f'/abc/alphabet/{arg}')
    assert r.status_code == 200
    assert isinstance(r.json(), str)

def test_arguments_route(dummy_project, application_domain):
    c = TestClient(application_domain)

    path = '/arguments'
    r = c.request('get', path)
    assert r.status_code == 400
    r = c.request('get', path, params={'foo':True})
    assert r.status_code == 400
    arg = {'foo':True, 'bar':True}
    r = c.request('get', path, params=arg)
    assert r.status_code == 200
    for key, val in arg.items():
        assert json.loads(r.json()[key]) == val
    path = '/async_router/arguments'
    r = c.request('get', path)
    assert r.status_code == 400
    r = c.request('get', path, params={'foo':True})
    assert r.status_code == 400
    arg = {'foo':True, 'bar':True}
    r = c.request('get', path, params=arg)
    assert r.status_code == 200
    for key, val in arg.items():
        assert json.loads(r.json()[key]) == val




import os
import jwt
from requests import Request
import pytest
from unittest.mock import patch
import json
from json.decoder import JSONDecodeError
import sys
from hashlib import sha256
from base64 import b64decode
from uuid import uuid4, UUID

from starlette.testclient import TestClient
from starlette.authentication import (
    AuthenticationBackend, AuthenticationError, BaseUser, AuthCredentials,
    UnauthenticatedUser)
from starlette.responses import PlainTextResponse

from halfapi.lib.jwt_middleware import (
    JWTUser, JWTAuthenticationBackend,
    JWTWebSocketAuthenticationBackend)


def test_JWTUser():
    uid = uuid4()
    token = '{}'
    payload = {}
    user = JWTUser(uid, token, payload)
    assert user.id == uid
    assert user.token == token
    assert user.payload == payload
    assert user.is_authenticated == True

def test_jwt_NoToken(dummy_app):
    async def test_route(request):
        assert isinstance(request.user, UnauthenticatedUser)
        return PlainTextResponse('ok')

    dummy_app.add_route('/test', test_route)
    test_client = TestClient(dummy_app)
    resp = test_client.request('get', '/test')
    assert resp.status_code == 200

def test_jwt_Token(dummy_app, token_builder):
    async def test_route(request):
        assert isinstance(request.user, JWTUser)
        print(request.scope['app'].debug)
        return PlainTextResponse('ok')

    dummy_app.add_route('/test', test_route)
    test_client = TestClient(dummy_app)

    resp = test_client.request('get', '/test',
        cookies={
            'Authorization': token_builder
        })
    assert resp.status_code == 200

    resp = test_client.request('get', '/test',
        headers={
            'Authorization': token_builder
        })
    assert resp.status_code == 200



def test_jwt_DebugFalse(dummy_app, token_debug_false_builder):
    async def test_route(request):
        assert isinstance(request.user, JWTUser)
        return PlainTextResponse('ok')

    dummy_app.add_route('/test', test_route)
    test_client = TestClient(dummy_app)

    resp = test_client.request('get', '/test',
        cookies={
            'Authorization': token_debug_false_builder
        })
    assert resp.status_code == 200

    resp = test_client.request('get', '/test',
        headers={
            'Authorization': token_debug_false_builder
        })
    assert resp.status_code == 200


def test_jwt_DebugTrue(dummy_app, token_debug_true_builder):
    """
    A debug token should return a 400 status code with a non debug app
    """
    async def test_route(request):
        return PlainTextResponse('ok')

    dummy_app.add_route('/test', test_route)
    test_client = TestClient(dummy_app)

    resp = test_client.request('get', '/test',
        cookies={
            'Authorization': token_debug_true_builder
        })
    assert resp.status_code == 400

    resp = test_client.request('get', '/test',
        headers={
            'Authorization': token_debug_true_builder
        })
    assert resp.status_code == 400


def test_jwt_DebugTrue_DebugApp(dummy_debug_app, token_debug_true_builder):
    """
    A debug token should return a 200 status code with a debug app
    """
    async def test_route(request):
        assert isinstance(request.user, JWTUser)
        return PlainTextResponse('ok')

    dummy_debug_app.add_route('/test', test_route)
    test_client = TestClient(dummy_debug_app)

    resp = test_client.request('get', '/test',
        cookies={
              'Authorization': token_debug_true_builder
        })
    assert resp.status_code == 200

    resp = test_client.request('get', '/test',
        headers={
              'Authorization': token_debug_true_builder
        })
    assert resp.status_code == 200



#!/usr/bin/env python3
"""
cli/init.py Defines the "halfapi init" cli commands

Helps the user to create a new project
"""
# builtins
import os
import sys
import re

import click

from .. import __version__
from ..conf import CONF_DIR

from .cli import cli

from ..logging import logger

TMPL_HALFAPI_ETC = """[project]
host = 127.0.0.1
port = 8000
secret = /path/to/secret_file
production = False
base_dir = {base_dir}
"""

TMPL_HALFAPI_CONFIG = """[project]
halfapi_version = {halfapi_version}

[domain]
"""

@click.argument('project')
@cli.command()
def init(project):
    """
    The "halfapi init" command
    """
    if not re.match('^[a-z0-9_]+$', project, re.I):
        click.echo('Project name must match "^[a-z0-9_]+$", retry.', err=True)
        sys.exit(1)

    if os.path.exists(project):
        click.echo(f'A file named {project} already exists, abort.', err=True)
        sys.exit(1)


    logger.debug('Create directory %s', project)
    os.mkdir(project)

    logger.debug('Create directory %s/.halfapi', project)
    os.mkdir(f'{project}/.halfapi')

    with open(f'{project}/.halfapi/config', 'w') as conf_file:
        conf_file.write(TMPL_HALFAPI_CONFIG.format(
            halfapi_version=__version__))


    click.echo(f'Configure halfapi project in {CONF_DIR}/{project}')



import importlib
import inspect
import os
import re

from packaging.specifiers import SpecifierSet
from packaging.version import Version
from typing import Coroutine, Dict, Iterator, List, Tuple
from types import ModuleType, FunctionType

from schema import SchemaError

from starlette.applications import Starlette
from starlette.routing import Router, Route

import yaml


from . import __version__
from .lib.constants import API_SCHEMA_DICT, ROUTER_SCHEMA, VERBS
from .half_route import HalfRoute
from .lib import acl as lib_acl
from .lib.responses import PlainTextResponse
from .lib.routes import JSONRoute
from .lib.domain import MissingAclError, PathError, UnknownPathParameterType, \
    UndefinedRoute, UndefinedFunction, get_fct_name, route_decorator
from .lib.domain_middleware import DomainMiddleware
from .logging import logger

class HalfDomain(Starlette):
    def __init__(self, domain, module=None, router=None, acl=None, app=None):
        """
        Parameters:
            domain (str): Module name (should be importable)
            router (str): Router name (should be importable from domain module
                defaults to __router__ variable from domain module)
            app (HalfAPI): The app instance
        """
        self.app = app

        self.m_domain = importlib.import_module(domain) if module is None else module
        d_domain = getattr(self.m_domain, 'domain', domain)
        self.name = d_domain['name']
        self.id = d_domain['id']
        self.version = d_domain['version']
        self.halfapi_version = d_domain.get('halfapi_version', __version__)
        self.deps = d_domain.get('deps', tuple())

        if not router:
            self.router = d_domain.get('routers', '.routers')
        else:
            self.router = router

        self.m_router = None
        try:
            self.m_router = importlib.import_module(self.router, self.m_domain.__package__)
        except AttributeError:
            raise Exception('no router module')

        self.m_acl = HalfDomain.m_acl(self.m_domain, acl)

        self.config = { **app.config }

        logger.info('HalfDomain creation %s %s', domain, self.config)

        for elt in self.deps:
            package, version = elt
            specifier = SpecifierSet(version)
            package_module = importlib.import_module(package)
            if Version(package_module.__version__) not in specifier:
                raise Exception(
                    'Wrong version for package {} version {} (excepting {})'.format(
                        package, package_module.__version__, specifier
                    ))


        super().__init__(
            routes=self.gen_domain_routes(),
            middleware=[
                (DomainMiddleware, {
                    'domain': {
                        'name': self.name,
                        'id': self.id,
                        'version': self.version,
                        'halfapi_version': self.halfapi_version,
                        'config': self.config.get('domain', {}).get(self.name, {}).get('config', {})
                    }
                })
            ]
        )

    @staticmethod
    def name(module):
        """ Returns the name declared in the 'domain' dict at the root of the package
        """
        return module.domain['name']
        

    @staticmethod
    def m_acl(module, acl=None):
        """ Returns the imported acl module for the domain module
        """
        if not acl:
            acl = getattr(module, '__acl__', '.acl')

        return importlib.import_module(acl, module.__package__)


    @staticmethod
    def acls(module, acl=None):
        """ Returns the ACLS constant for the given domain
        """
        m_acl = HalfDomain.m_acl(module, acl)
        try:
            return [
                lib_acl.ACL(*elt)
                for elt in getattr(m_acl, 'ACLS')
            ]
        except AttributeError as exc:
            logger.error(exc)
            raise Exception(
                f'Missing acl.ACLS constant in module {m_acl.__package__}') from exc

    @staticmethod
    def acls_route(domain, module_path=None, acl=None):
        """ Dictionary of acls

        Format :

        {
            [acl_name]: {
                callable: fct_reference,
                docs: fct_docstring,
            }
        }
        """

        d_res = {}

        module = importlib.import_module(domain) \
            if module_path is None \
            else importlib.import_module(module_path)

        m_acl = HalfDomain.m_acl(module, acl)

        for elt in HalfDomain.acls(module, acl=acl):

            fct = getattr(m_acl, elt.name)

            d_res[elt.name] = {
                'callable': fct,
                'docs': elt.documentation
            }

        return d_res

    @staticmethod
    def acls_router(domain, module_path=None, acl=None):
        """ Router of the acls routes :

        / : Same result as HalfDomain.acls_route
        /acl_name : Dummy route protected by the "acl_name" acl
        """

        async def dummy_endpoint(request, *args, **kwargs):
            return PlainTextResponse('')

        routes = []
        d_res = {}

        module = importlib.import_module(domain) \
            if module_path is None \
            else importlib.import_module(module_path)


        m_acl = HalfDomain.m_acl(module, acl)

        for elt in HalfDomain.acls(module, acl=acl):

            fct = getattr(m_acl, elt.name)

            d_res[elt.name] = {
                'callable': fct,
                'docs': elt.documentation,
                'public': elt.public
            }

            if elt.public:
                routes.append(
                    Route(
                        f'/{elt.name}',
                        HalfRoute.acl_decorator(
                            dummy_endpoint,
                            params=[{'acl': fct}]
                        ),
                        methods=['GET']
                    )
                )

        d_res_under_domain_name = {}
        d_res_under_domain_name[HalfDomain.name(module)] = d_res

        routes.append(
            Route(
                '/',
                JSONRoute(d_res_under_domain_name),
                methods=['GET']
            )
        )

        return Router(routes)


    @staticmethod
    def gen_routes(m_router: ModuleType,
        verb: str,
        path: List[str],
        params: List[Dict]) -> Tuple[FunctionType, Dict]:
        """
        Returns a tuple of the function associatied to the verb and path arguments,
        and the dictionary of it's acls

        Parameters:
            - m_router (ModuleType): The module containing the function definition

            - verb (str): The HTTP verb for the route (GET, POST, ...)

            - path (List): The route path, as a list (each item being a level of
                deepness), from the lowest level (domain) to the highest

            - params (Dict): The acl list of the following format :
                [{'acl': Function, 'args': {'required': [], 'optional': []}}]


        Returns:

            (Function, Dict): The destination function and the acl dictionary

        """
        if len(params) == 0:
            raise MissingAclError('[{}] {}'.format(verb, '/'.join(path)))

        if len(path) == 0:
            logger.error('Empty path for [{%s}]', verb)
            raise PathError()

        fct_name = get_fct_name(verb, path[-1])
        if hasattr(m_router, fct_name):
            fct = getattr(m_router, fct_name)
        else:
            raise UndefinedFunction('{}.{}'.format(m_router.__name__, fct_name or ''))


        if not inspect.iscoroutinefunction(fct):
            return route_decorator(fct), params

        # TODO: Remove when using only sync functions
        return lib_acl.args_check(fct), params


    @staticmethod
    def gen_router_routes(m_router, path: List[str]) -> \
        Iterator[Tuple[str, str, ModuleType, Coroutine, List]]:
        """
        Recursive generator that parses a router (or a subrouter)
        and yields from gen_routes

        Parameters:

            - m_router (ModuleType): The currently treated router module
            - path (List[str]): The current path stack

        Yields:

            (str, str, ModuleType, Coroutine, List): A tuple containing the path, verb,
                router module, function reference and parameters of the route.
                Function and parameters are yielded from then gen_routes function,
                that decorates the endpoint function.
        """

        for subpath, params in HalfDomain.read_router(m_router).items():
            path.append(subpath)

            for verb in VERBS:
                if verb not in params:
                    continue
                yield ('/'.join(filter(lambda x: len(x) > 0, path)),
                    verb,
                    m_router,
                    *HalfDomain.gen_routes(m_router, verb, path, params[verb])
                )

            for subroute in params.get('SUBROUTES', []):
                #logger.debug('Processing subroute **%s** - %s', subroute, m_router.__name__)
                param_match = re.fullmatch('^([A-Z_]+)_([a-z]+)$', subroute)
                if param_match is not None:
                    try:
                        path.append('{{{}:{}}}'.format(
                            param_match.groups()[0].lower(),
                            param_match.groups()[1]))
                    except AssertionError as exc:
                        raise UnknownPathParameterType(subroute) from exc
                else:
                    path.append(subroute)

                try:
                    yield from HalfDomain.gen_router_routes(
                        importlib.import_module(f'.{subroute}', m_router.__name__),
                        path)

                except ImportError as exc:
                    logger.error('Failed to import subroute **{%s}**', subroute)
                    raise exc

                path.pop()

            path.pop()


    @staticmethod
    def read_router(m_router: ModuleType) -> Dict:
        """
        Reads a module and returns a router dict

        If the module has a "ROUTES" constant, it just returns this constant,
        Else, if the module has an "ACLS" constant, it builds the accurate dict

        TODO: May be another thing, may be not a part of halfAPI

        """
        m_path = None

        try:
            if not hasattr(m_router, 'ROUTES'):
                routes = {'':{}}
                acls = getattr(m_router, 'ACLS') if hasattr(m_router, 'ACLS') else None

                if acls is not None:
                    for method in acls.keys():
                        if method not in VERBS:
                            raise Exception(
                                'This method is not handled: {}'.format(method))

                        routes[''][method] = []
                        routes[''][method] = acls[method].copy()

                routes['']['SUBROUTES'] = []
                if hasattr(m_router, '__path__'):
                    """ Module is a package
                    """
                    m_path = getattr(m_router, '__path__')
                    if isinstance(m_path, list) and len(m_path) == 1:
                        routes['']['SUBROUTES'] = [
                            elt.name
                            for elt in os.scandir(m_path[0])
                            if elt.is_dir()
                        ]
            else:
                routes = getattr(m_router, 'ROUTES')

            try:
                ROUTER_SCHEMA.validate(routes)
            except SchemaError as exc:
                logger.error(routes)
                raise exc

            return routes
        except ImportError as exc:
            # TODO: Proper exception handling
            raise exc
        except FileNotFoundError as exc:
            # TODO: Proper exception handling
            logger.error(m_path)
            raise exc

    def gen_domain_routes(self):
        """
        Yields the Route objects for a domain

        Parameters:
            m_domains: ModuleType

        Returns:
            Generator(HalfRoute)
        """
        yield HalfRoute('/',
            JSONRoute([ self.schema() ]),
            [{'acl': lib_acl.public}],
            'GET'
        )

        for path, method, m_router, fct, params in HalfDomain.gen_router_routes(self.m_router, []):
            yield HalfRoute(f'/{path}', fct, params, method)

    def schema_dict(self) -> Dict:
        """ gen_router_routes return values as a dict
        Parameters:

        m_router (ModuleType): The domain routers' module

        Returns:

        Dict: Schema of dict is halfapi.lib.constants.DOMAIN_SCHEMA

        @TODO: Should be a "router_schema_dict" function
        """
        d_res = {}

        for path, verb, m_router, fct, parameters in HalfDomain.gen_router_routes(self.m_router, []):
            if path not in d_res:
                d_res[path] = {}

            if verb not in d_res[path]:
                d_res[path][verb] = {}

            d_res[path][verb]['callable'] = f'{m_router.__name__}:{fct.__name__}'
            try:
                d_res[path][verb]['docs'] = yaml.safe_load(fct.__doc__)
            except AttributeError:
                logger.error(
                    'Cannot read docstring from fct (fct=%s path=%s verb=%s', fct.__name__, path, verb)

            d_res[path][verb]['acls'] = list(map(lambda elt: { **elt, 'acl': elt['acl'].__name__ },
                parameters))

        return d_res


    def schema(self) -> Dict:
        schema = { **API_SCHEMA_DICT }
        schema['domain'] = {
            'name': self.name,
            'id': self.id,
            'version': getattr(self.m_domain, '__version__', ''),
            'patch_release': getattr(self.m_domain, '__patch_release__', ''),
            'routers': self.m_router.__name__,
            'acls': tuple(getattr(self.m_acl, 'ACLS', ()))
        }
        schema['paths'] = self.schema_dict()
        return schema



import importlib
import functools
import os
import sys
import json
from json.decoder import JSONDecodeError
from unittest import TestCase
from starlette.testclient import TestClient
from click.testing import CliRunner
from ..cli.cli import cli
from ..halfapi import HalfAPI
from ..half_domain import HalfDomain
from pprint import pprint
import tempfile

class TestDomain(TestCase):
    @property
    def domain_name(self):
        return getattr(self, 'DOMAIN')

    @property
    def module_name(self):
        return getattr(self, 'MODULE', self.domain_name)

    @property
    def acl_path(self):
        return getattr(self, 'ACL', '.acl')

    @property
    def router_path(self):
        return getattr(self, 'ROUTERS', '.routers')

    @property
    def router_module(self):
        return '.'.join((self.module_name, self.ROUTERS))


    def setUp(self):
        # CLI
        class_ = CliRunner
        def invoke_wrapper(f):
            """Augment CliRunner.invoke to emit its output to stdout.

            This enables pytest to show the output in its logs on test
            failures.

            """
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                echo = kwargs.pop('echo', False)
                result = f(*args, **kwargs)

                if echo is True:
                    sys.stdout.write(result.output)

                return result

            return wrapper

        class_.invoke = invoke_wrapper(class_.invoke)
        self.runner = class_(mix_stderr=False)

        # HTTP
        self.halfapi_conf = {
            'secret': 'testsecret',
            'production': False,
            'domain': {}
        }

        self.halfapi_conf['domain'][self.domain_name] = {
            'name': self.domain_name,
            'router': self.router_path,
            'acl': self.acl_path,
            'module': self.module_name,
            'prefix': False,
            'enabled': True,
            'config': getattr(self, 'CONFIG', {})
        }

        _, self.config_file = tempfile.mkstemp()
        with open(self.config_file, 'w') as fh:
            fh.write(json.dumps(self.halfapi_conf))

        self.halfapi = HalfAPI(self.halfapi_conf)

        self.client = TestClient(self.halfapi.application)

        self.module = importlib.import_module(
            self.module_name
        )


    def tearDown(self):
        pass

    def check_domain(self):
        result = None
        try:
            result = self.runner.invoke(cli, '--version')
            self.assertEqual(result.exit_code, 0)
            result = self.runner.invoke(cli, ['domain', '--read', self.DOMAIN, self.config_file])
            self.assertEqual(result.exit_code, 0)
            result_d = json.loads(result.stdout)
            result = self.runner.invoke(cli, ['run', '--help'])
            self.assertEqual(result.exit_code, 0)
            # result = self.runner.invoke(cli, ['run', '--dryrun', self.DOMAIN])
            # self.assertEqual(result.exit_code, 0)
        except AssertionError as exc:
            print(f'Result {result}')
            print(f'Stdout {result.stdout}')
            print(f'Stderr {result.stderr}')
            raise exc
        except JSONDecodeError as exc:
            print(f'Result {result}')
            print(f'Stdout {result.stdout}')
            raise exc



        return result_d

    def check_routes(self):
        r = self.client.request('get', '/')
        assert r.status_code == 200
        schemas = r.json()
        assert isinstance(schemas, list)
        for schema in schemas:
            assert isinstance(schema, dict)
            assert 'openapi' in schema
            assert 'info' in schema
            assert 'paths' in schema
            assert 'domain' in schema

        r = self.client.request('get', '/halfapi/acls')
        assert r.status_code == 200
        d_r = r.json()
        assert isinstance(d_r, dict)
        assert self.domain_name in d_r.keys()

        ACLS = HalfDomain.acls(self.module, self.acl_path)
        assert len(ACLS) == len(d_r[self.domain_name])

        for acl_rule in ACLS:
            assert len(acl_rule.name) > 0
            assert acl_rule.name in d_r[self.domain_name]
            assert len(acl_rule.documentation) > 0
            assert isinstance(acl_rule.priority, int)
            assert acl_rule.priority >= 0

            if acl_rule.public is True:
                r = self.client.request('get', f'/halfapi/acls/{acl_rule.name}')
                assert r.status_code in [200, 401]



#!/usr/bin/env python3
import pytest
from starlette.authentication import UnauthenticatedUser
from starlette.testclient import TestClient
import subprocess
import json
import os
import sys
import pprint
from halfapi.lib.constants import API_SCHEMA


def test_halfapi_whoami(application_debug):
    # @TODO : If we use isolated filesystem multiple times that creates a bug.
    # So we use a single function with fixture "application debug"

    c = TestClient(application_debug)
    r = c.request('get', '/halfapi/whoami')
    assert r.status_code == 200

def test_halfapi_log(application_debug):
    # @TODO : If we use isolated filesystem multiple times that creates a bug.
    # So we use a single function with fixture "application debug"

    c = TestClient(application_debug)

    r = c.request('get', '/halfapi/log')
    assert r.status_code == 200

def test_halfapi_error_400(application_debug):
    # @TODO : If we use isolated filesystem multiple times that creates a bug.
    # So we use a single function with fixture "application debug"

    c = TestClient(application_debug)

    r = c.request('get', '/halfapi/error/400')
    assert r.status_code == 400

def test_halfapi_error_404(application_debug):
    # @TODO : If we use isolated filesystem multiple times that creates a bug.
    # So we use a single function with fixture "application debug"

    c = TestClient(application_debug)

    r = c.request('get', '/halfapi/error/404')
    assert r.status_code == 404

def test_halfapi_error_500(application_debug):
    # @TODO : If we use isolated filesystem multiple times that creates a bug.
    # So we use a single function with fixture "application debug"

    c = TestClient(application_debug)

    r = c.request('get', '/halfapi/error/500')
    assert r.status_code == 500

def test_schema(application_debug):
    c = TestClient(application_debug)

    r = c.request('get', '/')
    schemas = r.json()
    assert isinstance(schemas, list)
    for schema in schemas:
        assert isinstance(schema, dict)
        assert API_SCHEMA.validate(schema)



"""
DomainMiddleware
"""
from starlette.datastructures import URL
from starlette.middleware.base import (BaseHTTPMiddleware,
    RequestResponseEndpoint)
from starlette.requests import Request
from starlette.responses import Response

from ..logging import logger

class DomainMiddleware(BaseHTTPMiddleware):
    """
    DomainMiddleware adds the api routes and acls to the following scope keys :

        - api
        - acl
    """

    def __init__(self, app, domain):
        """ app: HalfAPI instance
        """
        logger.info('DomainMiddleware app:%s domain:%s', app, domain)
        super().__init__(app)
        self.domain = domain
        self.request = None


    async def dispatch(self, request: Request,
        call_next: RequestResponseEndpoint) -> Response:
        """
        Call of the route fonction (decorated or not)
        """

        request.scope['domain'] = self.domain['name']
        if hasattr(request.app, 'config') \
          and isinstance(request.app.config, dict):
            # Set the config scope to the domain's config
            request.scope['config'] = request.app.config.get(
                'domain', {}
            ).get(
                self.domain['name'], {}
            ).copy()

            # TODO: Remove in 0.7.0
            config = request.scope['config'].copy()
            request.scope['config']['domain'] = {}
            request.scope['config']['domain'][self.domain['name']] = {}
            request.scope['config']['domain'][self.domain['name']]['config'] = config
            


        else:
            logger.debug('%s', request.app)
            logger.debug('%s', getattr(request.app, 'config', None))

        response = await call_next(request)

        if 'acl_pass' in request.scope:
            # Set the http header "x-acl" if an acl was used on the route
            response.headers['x-acl'] = request.scope['acl_pass']

        if 'args' in request.scope:
            # Set the http headers "x-args-required" and "x-args-optional"

            if len(request.scope['args'].get('required', set())):
                response.headers['x-args-required'] = \
                    ','.join(request.scope['args']['required'])
            if len(request.scope['args'].get('optional', set())):
                response.headers['x-args-optional'] = \
                    ','.join(request.scope['args']['optional'])

        if len(request.scope.get('out', set())):
            response.headers['x-out'] = \
                ','.join(request.scope['out'])


        response.headers['x-domain'] = self.domain['name']

        return response



"""
JWT Middleware module

Classes:
    - JWTUser : goes in request.user
    - JWTAuthenticationBackend
    - JWTWebSocketAuthenticationBackend

Raises:
    Exception: If configuration has no SECRET
"""

from os import environ
import typing
from uuid import UUID

from http.cookies import SimpleCookie
import jwt
from starlette.authentication import (
    AuthenticationBackend, AuthenticationError, BaseUser, AuthCredentials,
    UnauthenticatedUser)
from starlette.requests import HTTPConnection, Request
from starlette.exceptions import HTTPException

from .user import CheckUser, JWTUser, Nobody
from ..logging import logger
from ..conf import CONFIG
from ..lib.responses import ORJSONResponse

SECRET=None

try:
    with open(CONFIG.get('secret', ''), 'r') as secret_file:
        SECRET = secret_file.read().strip()
except FileNotFoundError:
        logger.error('Could not import SECRET variable from conf module,'\
            ' using HALFAPI_SECRET environment variable')

def cookies_from_scope(scope):
    cookie = dict(scope.get("headers") or {}).get(b"cookie")
    if not cookie:
        return {}

    simple_cookie = SimpleCookie()
    simple_cookie.load(cookie.decode("utf8"))
    return {key: morsel.value for key, morsel in simple_cookie.items()}

def on_auth_error(request: Request, exc: Exception):
    response = ORJSONResponse({"error": str(exc)}, status_code=401)
    response.delete_cookie('Authorization')
    return response

class JWTAuthenticationBackend(AuthenticationBackend):
    def __init__(self, secret_key: str = SECRET,
        algorithm: str = 'HS256', prefix: str = 'JWT'):

        if secret_key is None:
            raise Exception('Missing secret_key argument for JWTAuthenticationBackend')
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.prefix = prefix

    @property
    def id(self) -> str:
        return self.__id

    async def authenticate(
        self, conn: HTTPConnection
    ) -> typing.Optional[typing.Tuple['AuthCredentials', 'BaseUser']]:

        # Standard way to authenticate via API
        # https://datatracker.ietf.org/doc/html/rfc7235#section-4.2
        token = conn.headers.get('Authorization')

        if not token:
            token = cookies_from_scope(conn.scope).get('Authorization')

        is_check_call = 'check' in conn.query_params

        PRODUCTION = conn.scope['app'].debug == False

        if not token and not is_check_call:
            return AuthCredentials(), Nobody()

        try:
            if token:
                payload = jwt.decode(token,
                    key=self.secret_key,
                    algorithms=[self.algorithm],
                    options={
                        'verify_signature': True
                    })

            if is_check_call:
                if token:
                    return AuthCredentials(), CheckUser(payload['user_id'])

                return AuthCredentials(), Nobody()


            if PRODUCTION and 'debug' in payload.keys() and payload['debug']:
                raise AuthenticationError(
                    'Trying to connect using *DEBUG* token in *PRODUCTION* mode')

        except jwt.ExpiredSignatureError as exc:
            return AuthCredentials(), Nobody()
        except jwt.InvalidTokenError as exc:
            raise AuthenticationError(str(exc)) from exc
        except Exception as exc:
            logger.error('Authentication error : %s', exc)
            raise exc


        return AuthCredentials(["authenticated"]), JWTUser(
            user_id=payload['user_id'], token=token, payload=payload)



class JWTWebSocketAuthenticationBackend(AuthenticationBackend):

    def __init__(self, secret_key: str, algorithm: str = 'HS256', query_param_name: str = 'jwt',
                 user_id: UUID = None, audience = None):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.query_param_name = query_param_name
        self.__id = user_id
        self.audience = audience


    async def authenticate(
        self, conn: HTTPConnection
    ) -> typing.Optional[typing.Tuple["AuthCredentials", "BaseUser"]]:

        if self.query_param_name not in conn.query_params:
            return AuthCredentials(), Nobody()

        token = conn.query_params[self.query_param_name]

        try:
            payload = jwt.decode(
                token,
                key=self.secret_key,
                algorithms=[self.algorithm],
                audience=self.audience,
                options={
                    'verify_signature': bool(PRODUCTION)
                })

            if PRODUCTION and 'debug' in payload.keys() and payload['debug']:
                raise AuthenticationError(
                    'Trying to connect using *DEBUG* token in *PRODUCTION* mode')

        except jwt.InvalidTokenError as exc:
            raise AuthenticationError(str(exc)) from exc

        return (
            AuthCredentials(["authenticated"]),
            JWTUser(
                user_id=payload['id'],
                token=token,
                payload=payload)
        )



import json
import decimal
import datetime

from starlette.responses import Response
from halfapi.lib.responses import *

def test_orjson():
    test_obj = {
        "ok": "ko",
        "dec": decimal.Decimal(42),
        "set": {0, 4, 2},
        "date": datetime.date(1,1,1),
        "datetime": datetime.datetime(1,1,1),
    }

    resp = ORJSONResponse(test_obj)
    body = resp.body.decode()
    test_obj_dec = json.loads(body)
    print(test_obj_dec)
    assert 'ok' in test_obj_dec.keys()
    assert isinstance(test_obj_dec['ok'], str)
    assert isinstance(test_obj_dec['dec'], str)
    assert isinstance(test_obj_dec['set'], list)
    assert isinstance(test_obj_dec['date'], str)
    assert test_obj_dec['date'] == '0001-01-01'
    assert test_obj_dec['datetime'] == '0001-01-01T00:00:00'


def test_responses():
    resp = HJSONResponse('')
    assert isinstance(resp, Response)
    assert resp.status_code == 200

    resp = ORJSONResponse('')
    assert isinstance(resp, Response)
    assert resp.status_code == 200

    resp = PlainTextResponse()
    assert isinstance(resp, Response)
    assert resp.status_code == 200

def test_errors():
    resp = ServiceUnavailableResponse()
    assert isinstance(resp, Response)
    assert resp.status_code == 503

    resp = UnauthorizedResponse()
    assert isinstance(resp, Response)
    assert resp.status_code == 401

    resp = InternalServerErrorResponse()
    assert isinstance(resp, Response)
    assert resp.status_code == 500

    resp = NotFoundResponse()
    assert isinstance(resp, Response)
    assert resp.status_code == 404

    resp = NotImplementedResponse()
    assert isinstance(resp, Response)
    assert resp.status_code == 501



#!/usr/bin/env python3
"""
This is the *query* library that contains all the useful functions to treat our
queries

Fonction:
    - parse_query
"""

from starlette.exceptions import HTTPException


def parse_query(q_string: str = ""):
    """
    Returns the fitting Response object according to query parameters.

    The parse_query function handles the following arguments in the query
    string : format, limit, and offset
    It returns a callable function that returns the desired Response object.

        Parameters:
            q_string (str): The query string "q" parameter, in the format
                key0:value0|...|keyN:valueN

        Returns:
            Callable[[half_orm.model.Model], Response]

        Available query arguments:
            format:
                - csv
                - json
            limit: int > 0
            offset: int > 0


        Examples:

            >>> parse_query()
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('format:csv')
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('format:json')
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('format:csv|limit:10')
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('format:csv|offset:10')
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('format:csv|limit:10|offset:10')
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('limit:10')
            <function parse_query.<locals>.select at 0x...>

            >>> parse_query('limit=10')
            Traceback (most recent call last):
                ...
            starlette.exceptions.HTTPException: 400


    """

    params = {}
    if len(q_string) > 0:
        try:
            split_ = lambda x : x.split(':')
            params = dict(map(split_, q_string.split('|')))
        except ValueError as exc:
            raise HTTPException(400) from exc
        split_ = lambda x : x.split(':')
        params = dict(map(split_, q_string.split('|')))

    def select(obj, fields):

        if 'limit' in params and int(params['limit']) > 0:
            obj.limit(int(params['limit']))

        if 'offset' in params and int(params['offset']) > 0:
            obj.offset(int(params['offset']))

        return list(obj.select(*fields))

    return select



#!/usr/bin/env python3
"""
cli/config.py Contains the .halfapi/config

Defines the "halfapi config" command
"""
import click

from .cli import cli
from ..conf import CONFIG

DOMAIN_CONF_STR="""
[domain]
name = {name}
router = {router}
"""

CONF_STR="""
[project]
host = {host}
port = {port}
production = {production}
"""


@cli.command()
def config():
    """
    Lists config parameters and their values
    """
    click.echo(CONF_STR.format(**CONFIG))



#!/usr/bin/env python3
"""
app.py is the file that is read when launching the application using an asgi
runner.

It defines the following globals :

    - routes (contains the Route objects for the application)
    - application (the asgi application itself - a starlette object)

"""
import logging
import time
import importlib
from datetime import datetime

# asgi framework
from starlette.applications import Starlette
from starlette.authentication import UnauthenticatedUser
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.routing import Router, Route, Mount
from starlette.requests import Request
from starlette.responses import Response, PlainTextResponse
from starlette.middleware.authentication import AuthenticationMiddleware

from timing_asgi import TimingMiddleware
from timing_asgi.integrations import StarletteScopeToName

# module libraries

from .lib.constants import API_SCHEMA_DICT
from .lib.domain_middleware import DomainMiddleware
from .lib.timing import HTimingClient
from .lib.jwt_middleware import JWTAuthenticationBackend, on_auth_error
from .lib.responses import (ORJSONResponse, UnauthorizedResponse,
    NotFoundResponse, InternalServerErrorResponse, NotImplementedResponse,
    ServiceUnavailableResponse, gen_exception_route)
from .lib.domain import NoDomainsException
from .lib.routes import gen_schema_routes, JSONRoute
from .lib.schemas import schema_json
from .logging import logger, config_logging
from .half_domain import HalfDomain
from halfapi import __version__

class HalfAPI(Starlette):
    def __init__(self, config,
        d_routes=None):
        config_logging(logging.DEBUG)
        self.config = config

        SECRET = self.config.get('secret')
        PRODUCTION = self.config.get('production', True)
        DRYRUN = self.config.get('dryrun', False)

        self.PRODUCTION = PRODUCTION
        self.SECRET = SECRET

        # Domains

        """ HalfAPI routes (if not PRODUCTION, includes debug routes)
        """
        routes = []
        routes.append(
            Mount('/halfapi', routes=list(self.halfapi_routes()))
        )

        logger.info('Config: %s', self.config)

        domains = {
            key: elt
            for key, elt in self.config.get('domain', {}).items()
            if elt.get('enabled', False)
        }

        logger.info('Active domains: %s', domains)

        if d_routes:
            # Mount the routes from the d_routes argument - domain-less mode
            logger.info('Domain-less mode : the given schema defines the activated routes')
            for route in gen_schema_routes(d_routes):
                routes.append(route)
        else:
            pass

        startup_fcts = []

        if DRYRUN:
            startup_fcts.append(
                HalfAPI.wait_quit()
            )

        super().__init__(
            debug=not PRODUCTION,
            routes=routes,
            exception_handlers={
                401: gen_exception_route(UnauthorizedResponse),
                404: gen_exception_route(NotFoundResponse),
                500: gen_exception_route(HalfAPI.exception),
                501: gen_exception_route(NotImplementedResponse),
                503: gen_exception_route(ServiceUnavailableResponse)
            },
            on_startup=startup_fcts
        )

        schemas = []

        self.__domains = {}

        for key, domain in domains.items():
            if not isinstance(domain, dict):
                continue

            dom_name = domain.get('name', key)
            if not domain.get('enabled', False):
                continue

            if not domain.get('prefix', False):
                if len(domains.keys()) > 1:
                    raise Exception('Cannot use multiple domains and set prefix to false')
                path = '/'
            else:
                path = f'/{dom_name}'

            logger.debug('Mounting domain %s on %s', domain.get('name'), path)

            domain_key = domain.get('name', key)

            add_domain_args = {
                **domain,
                'path': path
            }

            self.add_domain(**add_domain_args)

            schemas.append(self.__domains[domain_key].schema())


        self.add_route('/', JSONRoute(schemas))

        if SECRET:
            self.add_middleware(
                AuthenticationMiddleware,
                backend=JWTAuthenticationBackend(),
                on_error=on_auth_error
            )

        if not PRODUCTION:
            self.add_middleware(
                TimingMiddleware,
                client=HTimingClient(),
                metric_namer=StarletteScopeToName(prefix="halfapi",
                starlette_app=self)
            )

    @property
    def version(self):
        return __version__

    async def version_async(self, request, *args, **kwargs):
        return Response(self.version)

    @staticmethod
    async def exception(request: Request, exc: HTTPException):
        logger.critical(exc, exc_info=True)
        return InternalServerErrorResponse()

    @property
    def application(self):
        return self

    def halfapi_routes(self):
        """ Halfapi default routes
        """

        async def get_user(request, *args, **kwargs):
            return ORJSONResponse({'user':request.user})

        yield Route('/whoami', get_user)
        yield Route('/schema', schema_json)
        yield Mount('/acls', self.acls_router())
        yield Route('/version', self.version_async)
        """ Halfapi debug routes definition
        """
        if self.PRODUCTION:
            return

        """ Debug routes
        """
        async def debug_log(request: Request, *args, **kwargs):
            logger.debug('debuglog# %s', {datetime.now().isoformat()})
            logger.info('debuglog# %s', {datetime.now().isoformat()})
            logger.warning('debuglog# %s', {datetime.now().isoformat()})
            logger.error('debuglog# %s', {datetime.now().isoformat()})
            logger.critical('debuglog# %s', {datetime.now().isoformat()})
            return Response('')
        yield Route('/log', debug_log)

        async def error_code(request: Request, *args, **kwargs):
            code = request.path_params['code']
            raise HTTPException(code)

        yield Route('/error/{code:int}', error_code)

        async def exception(request: Request, *args, **kwargs):
            raise Exception('Test exception')

        yield Route('/exception', exception)

    @staticmethod
    def api_schema(domain):
        pass

    @staticmethod
    def wait_quit():
        """ sleeps 1 second and quits. used in dry-run mode
        """
        import time
        import sys
        time.sleep(1)
        sys.exit(0)

    def acls_router(self):
        mounts = {}

        for domain, domain_conf in self.config.get('domain', {}).items():
            if isinstance(domain_conf, dict) and domain_conf.get('enabled', False):
                mounts['domain'] = HalfDomain.acls_router(
                    domain,
                    module_path=domain_conf.get('module'),
                    acl=domain_conf.get('acl')
                )

        if len(mounts) > 1:
            return Router([
                Mount(f'/{domain}', acls_router)
                for domain, acls_router in mounts.items()
            ])
        elif len(mounts) == 1:
            return Mount('/', mounts.popitem()[1])
        else:
            return Router()

    @property
    def domains(self):
        return self.__domains

    def add_domain(self, **kwargs):

        if not kwargs.get('enabled'):
            raise Exception(f'Domain not enabled ({kwargs})')

        name = kwargs['name']

        self.config['domain'][name] = kwargs.get('config', {})

        if not kwargs.get('module'):
            module = name
        else:
            module = kwargs.get('module')

        try:
            self.__domains[name] = HalfDomain(
                name,
                module=importlib.import_module(module),
                router=kwargs.get('router'),
                acl=kwargs.get('acl'),
                app=self
            )

        except ImportError as exc:
            print(
                'Cannot instantiate HalfDomain {} with module {}'.format(
                name,
                module
            ))
            raise exc

        self.mount(kwargs.get('path', name), self.__domains[name])

        return self.__domains[name]



from starlette.testclient import TestClient
from starlette.middleware.base import BaseHTTPMiddleware
from unittest.mock import patch
from halfapi.lib.domain_middleware import DomainMiddleware

def test_init():
    with patch('starlette.middleware.base.BaseHTTPMiddleware.__init__') as init:
        mw = DomainMiddleware('app', 'domain')
        init.assert_called_once_with('app')
        assert mw.domain == 'domain'
        assert mw.request == None

def test_call(application_debug):
    c = TestClient(application_debug)
    r = c.request('get', '/abc/alphabet')
    assert r.status_code == 200
    assert r.headers['x-domain'] == 'dummy_domain'
    assert r.headers['x-acl'] == 'public'

    r = c.request('get', '/arguments')
    assert r.status_code == 400
    assert r.headers['x-domain'] == 'dummy_domain'
    assert r.headers['x-acl'] == 'public'
    assert 'foo' in r.headers['x-args-required'].split(',')
    assert 'bar' in r.headers['x-args-required'].split(',')
    assert r.headers['x-args-optional'] == 'x'

    c = TestClient(application_debug)
    r = c.request('post', '/arguments')
    assert r.status_code == 400
    assert r.headers['x-domain'] == 'dummy_domain'
    assert r.headers['x-acl'] == 'public'
    assert 'foo' in r.headers['x-args-required'].split(',')
    assert 'baz' in r.headers['x-args-required'].split(',')
    assert 'truebidoo' in r.headers['x-args-optional'].split(',')
    assert 'z' in r.headers['x-args-optional'].split(',')





import importlib
from halfapi.testing.test_domain import TestDomain

def test_dummy_domain():
    from . import dummy_domain
    from .dummy_domain import acl
    assert acl.public() is True
    assert isinstance(acl.random(), int)
    assert acl.private() is False


    from .dummy_domain import routers
    from .dummy_domain.routers.arguments import get
    from .dummy_domain.routers.abc.alphabet.TEST_uuid import get
    from .dummy_domain.routers.abc.pinnochio import get
    from .dummy_domain.routers.config import get
    from .dummy_domain.routers.config import get
    from .dummy_domain.routers import async_router
    from .dummy_domain.routers.async_router import ROUTES, get_abc_alphabet_TEST, get_abc_pinnochio, get_config, get_arguments



from schema import Schema
def test_constants():
    from halfapi.lib.constants import (
        VERBS,
        ACLS_SCHEMA,
        ROUTE_SCHEMA,
        DOMAIN_SCHEMA,
        API_SCHEMA)

    assert isinstance(VERBS, tuple)
    assert isinstance(ACLS_SCHEMA, Schema)
    assert isinstance(ROUTE_SCHEMA, Schema)
    assert isinstance(DOMAIN_SCHEMA, Schema)
    assert isinstance(API_SCHEMA, Schema)



import os
from starlette.applications import Starlette
from unittest.mock import MagicMock, patch
from halfapi.halfapi import HalfAPI

from halfapi.lib.domain import NoDomainsException

def test_halfapi_dummy_domain(dummy_domain):
    with patch('starlette.applications.Starlette') as mock:
        mock.return_value = MagicMock()
        config = {}
        config['domain'] = {}
        config['domain'][dummy_domain['name']] = dummy_domain
        print(config)
        halfapi = HalfAPI(config)



#!/usr/bin/env python3
"""
cli/domain.py Defines the "halfapi run" cli command
"""
import os
import sys
import click
import uvicorn

from .cli import cli
from .domain import list_api_routes
from ..conf import CONFIG, SCHEMA
from ..logging import logger
from ..lib.schemas import schema_csv_dict
from ..half_domain import HalfDomain

@click.option('--host', default=CONFIG.get('host'))
@click.option('--port', default=CONFIG.get('port'))
@click.option('--reload', default=False)
@click.option('--secret', default=CONFIG.get('secret'))
@click.option('--production', default=CONFIG.get('secret'))
@click.option('--loglevel', default=CONFIG.get('loglevel'))
@click.option('--prefix', default='/')
@click.option('--check', default=True)
@click.option('--dryrun', default=False, is_flag=True)
@click.argument('schema', type=click.File('r'), required=False)
@click.argument('domain', required=False)
@cli.command()
def run(host, port, reload, secret, production, loglevel, prefix, check, dryrun,
        schema, domain):
    """
    The "halfapi run" command
    """
    logger.debug('[run] host=%s port=%s reload=%s secret=%s production=%s loglevel=%s prefix=%s schema=%s',
        host, port, reload, secret, production, loglevel, prefix, schema
    )

    port = int(port)

    if production and reload:
        reload = False
        raise Exception('Can\'t use live code reload in production')

    click.echo(f'Launching application')

    if secret:
        CONFIG['secret'] = secret

    if schema:
        # Populate the SCHEMA global with the data from the given file
        for key, val in schema_csv_dict(schema, prefix).items():
            SCHEMA[key] = val

    if domain:
        # If we specify a domain to run as argument
        if 'domain' not in CONFIG:
            CONFIG['domain'] = {}

        # Disable all domains
        keys = list(CONFIG.get('domain').keys())
        for key in keys:
            CONFIG['domain'].pop(key)

        # And activate the desired one, mounted without prefix
        CONFIG['domain'][domain] = {
            'name': domain,
            'prefix': False,
            'enabled': True
        }

    # list_api_routes()

    click.echo(f'uvicorn.run("halfapi.app:application"\n' \
        f'host: {host}\n' \
        f'port: {port}\n' \
        f'log_level: {loglevel}\n' \
        f'reload: {reload}\n'
    )

    if dryrun:
        CONFIG['dryrun'] = True

    uvicorn.run('halfapi.app:application',
        host=host,
        port=int(port),
        log_level=loglevel,
        reload=reload)



import re
from schema import Schema, Optional, Or
from .. import __version__

VERBS = ('GET', 'POST', 'PUT', 'PATCH', 'DELETE')

ITERABLE_STR = Or([ str ], { str }, ( str ))

ACLS_SCHEMA = Schema([{
    'acl': str,
    Optional('args'): {
        Optional('required'): ITERABLE_STR,
        Optional('optional'): ITERABLE_STR
    },
    Optional('out'): ITERABLE_STR
}])
ROUTER_ACLS_SCHEMA = Schema([{
    'acl': lambda n: callable(n),
    Optional('args'): {
        Optional('required'): ITERABLE_STR,
        Optional('optional'): ITERABLE_STR
    },
    Optional('out'): ITERABLE_STR
}])


is_callable_dotted_notation = lambda x: re.match(
    r'^(([a-zA-Z_])+\.?)*:[a-zA-Z_]+$', 'ab_c.TEST:get')

ROUTE_SCHEMA = Schema({
    Optional(str): { # path - Optional when no routes
        str: { # method
            'callable': is_callable_dotted_notation,
            'docs': lambda n: True, # Should validate an openAPI spec
            'acls': ACLS_SCHEMA
        }
    }
})

DOMAIN_SCHEMA = Schema({
    'name': str,
    'id': str,
    Optional('routers'): str,
    Optional('version'): str,
    Optional('patch_release'): str,
    Optional('acls'): [
        [str, str, int, Optional(bool)]
    ]
})

API_SCHEMA_DICT = {
    'openapi': '3.0.0',
    'info': {
        'title': 'HalfAPI',
        'version': __version__
    },
}

API_SCHEMA = Schema({
    **API_SCHEMA_DICT,
    'domain': DOMAIN_SCHEMA,
    'paths': ROUTE_SCHEMA
})

ROUTER_SCHEMA = Schema({
    Or('', str): {
        # Optional('GET'): [],#ACLS_SCHEMA,
        Optional(Or(*VERBS)): ROUTER_ACLS_SCHEMA,
        Optional('SUBROUTES'): [Optional(str)]
    }
})



#!/usr/bin/env python3
"""
cli/domain.py Defines the "halfapi domain" cli commands
"""
# builtins
import os
import sys
import importlib
import subprocess

import json

import click
import orjson
import uvicorn


from .cli import cli
from ..conf import CONFIG

from ..half_domain import HalfDomain

from ..lib.routes import api_routes
from ..lib.responses import ORJSONResponse


from ..logging import logger

#################
# domain create #
#################
def create_domain(domain_name: str, module_path: str):
    logger.info('Will add **%s** (%s) to current halfAPI project',
            domain_name, module_path)

    #if domain_name in DOMAINSDICT():
    #    logger.warning('Domain **%s** is already in project', domain_name)
    #    sys.exit(1)

    def domain_tree_create():
        def create_init(path):
            with open(os.path.join(os.getcwd(), path, '__init__.py'), 'w') as f:
                f.writelines([
                    '"""',
                    f'name: {domain_name}',
                    f'router: {module_path}',
                    '"""'
                ])
                logger.debug('created %s', os.path.join(os.getcwd(), path, '__init__.py'))

        def create_acl(path):
            with open(os.path.join(path, 'acl.py'), 'w') as f:
                f.writelines([
                    'from halfapi.lib.acl import public, ACLS',

                ])
                logger.debug('created %s', os.path.join(path, 'acl.py'))


        os.mkdir(domain_name)
        create_init(domain_name)
        router_path = os.path.join(domain_name, 'routers')
        create_acl(domain_name)
        os.mkdir(router_path)
        create_init(router_path)

    # TODO: Generate config file

    domain_tree_create()
    """
    try:
        importlib.import_module(module_path)
    except ImportError:
        logger.error('cannot import %s', domain_name)
        domain_tree_create()
    """

    """
    try:
        importlib.import_module(domain_name)
    except ImportError:
        click.echo('Error in domain creation')
        logger.debug('%s', subprocess.run(['tree', 'a', os.getcwd()]))
        raise Exception('cannot create domain {}'.format(domain_name))
    """



###############
# domain read #
###############
def list_routes(domain, m_dom):
    """
    Echoes the list of the **m_dom** active routes
    """

    click.echo(f'\nDomain : {domain}\n')

    routes = api_routes(m_dom)[0]
    if len(routes):
        for key, item in routes.items():
            methods = '|'.join(list(item.keys()))
            click.echo(f'\t{key} : {methods}')
    else:
        click.echo('\t**No ROUTES**')
        raise Exception('Routeless domain')



def list_api_routes():
    """
    Echoes the list of all active domains.

    TODO: Rewrite function
    """

    click.echo('# API Routes')
    # for domain, m_dom in DOMAINSDICT().items():
    #     list_routes(domain, m_dom)


@click.option('--dry-run',default=False, is_flag=True)
@click.option('--run',default=False, is_flag=True)
@click.option('--read',default=False, is_flag=True)
@click.option('--create',default=False, is_flag=True)
@click.option('--update',default=False, is_flag=True)
@click.option('--delete',default=False, is_flag=True)
@click.argument('config_file', type=click.File(mode='rb'), required=False)
@click.argument('domain',default=None, required=False)
@cli.command()
def domain(domain, config_file, delete, update, create, read, run, dry_run):  #, domains, read, create, update, delete):
    """
    The "halfapi domain" command

    Parameters:
        domain (str|None): The domain name

        The parameter has a misleading name as it is a multiple option
        but this would be strange to use it several times named as "domains"

        update (boolean): If set, update the database for the selected domains
    """
    if not domain:
        if create:
            # TODO: Connect to the create_domain function
            raise NotImplementedError
        raise Exception('Missing domain name')
    if create:
        raise NotImplementedError
    elif update:
        raise NotImplementedError
    elif delete:
        raise NotImplementedError
    elif read:
        from ..conf import CONFIG
        from ..halfapi import HalfAPI

        if config_file:
            CONFIG = json.loads(''.join(
                [ line.decode() for line in config_file.readlines() ]
            ))

        halfapi = HalfAPI(CONFIG)
        click.echo(orjson.dumps(
            halfapi.domains[domain].schema(),
            option=orjson.OPT_NON_STR_KEYS,
            default=ORJSONResponse.default_cast)
        )

    else:
        from ..conf import CONFIG
        if 'domain' not in CONFIG:
            CONFIG['domain'] = {}

        if domain not in CONFIG['domain']:
            CONFIG['domain'][domain] = {
                'enabled': True,
                'name': domain
            }

        if dry_run:
            CONFIG['dryrun'] = True

        CONFIG['domain'][domain]['enabled'] = True
        port = CONFIG['domain'][domain].get('port', 3000)

        uvicorn.run(
            'halfapi.app:application',
            port=port
        )


    sys.exit(0)



import pytest
from halfapi.testing.test_domain import TestDomain
from pprint import pprint

class TestDummyDomain(TestDomain):
    from .dummy_domain import domain
    __name__ = domain.get('name')
    __routers__ = domain.get('routers')

    DOMAIN = __name__
    CONFIG = {'test': True}

    def test_domain(self):
        self.check_domain()

    def test_routes(self):
        self.check_routes()

    def test_html_route(self):
        res = self.client.request('get', '/ret_type')
        assert res.status_code == 200
        assert isinstance(res.content.decode(), str)
        assert res.headers['content-type'].split(';')[0] == 'text/html'

        res = self.client.request('get', '/ret_type/h24')
        assert res.status_code == 200
        assert isinstance(res.content.decode(), str)
        assert res.headers['content-type'].split(';')[0] == 'text/html'

        res = self.client.request('get', '/ret_type/h24/config')
        assert res.status_code == 200
        assert isinstance(res.content.decode(), str)
        assert res.headers['content-type'].split(';')[0] == 'text/html'

        res = self.client.request('post', '/ret_type/h24/config', json={
            'trou': 'glet'
        })
        assert res.status_code == 200
        assert isinstance(res.content.decode(), str)
        assert res.headers['content-type'].split(';')[0] == 'text/html'

    def test_arguments__get_routes(self):
        res = self.client.request('post', '/arguments?foo=1&x=3')

        assert res.status_code == 400

        arg_dict = {'foo': '1', 'bar': '2', 'x': '3'}
        res = self.client.request('get', '/arguments?foo=1&bar=2&x=3')
        assert res.json() == arg_dict

        res = self.client.request('get', '/arguments?foo=1&bar=2&x=3&y=4')
        assert res.json() == arg_dict

    def test_arguments_post_routes(self):
        arg_dict = {}
        res = self.client.request('post', '/arguments', json=arg_dict)

        assert res.status_code == 400

        arg_dict = {'foo': '1', 'bar': '3'}
        res = self.client.request('post', '/arguments', json=arg_dict)

        assert res.status_code == 400

        arg_dict = {'foo': '1', 'baz': '3'}
        res = self.client.request('post', '/arguments', json=arg_dict)

        assert res.json() == arg_dict

        arg_dict = {'foo': '1', 'baz': '3', 'truebidoo': '4'}
        res = self.client.request('post', '/arguments', json=arg_dict)

        assert res.json() == arg_dict

        res = self.client.request('post', '/arguments', json={ **arg_dict, 'y': '4'})
        assert res.json() == arg_dict

        res = self.client.request('post', '/arguments', json={ **arg_dict, 'z': True})
        assert res.json() == {**arg_dict, 'z': True}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re

from setuptools import setup, find_packages

import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

module_name="halfapi"
setup(
    name=module_name,
    version=get_version(module_name),
    url="https://github.com/halfAPI/halfapi",
    description="Core to write deep APIs using a module's tree",
    author="Maxime ALVES",
    author_email="maxime@freepoteries.fr",
    license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=get_packages(module_name),
    python_requires=">=3.8",
    install_requires=[
        "PyJWT>=2.6.0,<2.7.0",
        "starlette>=0.23,<0.24",
        "click>=8,<9",
        "uvicorn>=0.13,<1",
        "orjson>=3.8.5,<4",
        "pyyaml>=6,<7",
        "timing-asgi>=0.2.1,<1",
        "schema>=0.7.4,<1",
        "toml>=0.10,<0.11",
        "packaging>=19.0",
        "python-multipart"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
    ],
    extras_require={
        "tests":[
            "pytest>=7,<8",
            "pytest-asyncio",
            "pylint",
            "requests",
            "httpx"
        ],
        "pyexcel":[
            "pyexcel",
            "pyexcel-ods3",
            "pyexcel-xlsx"
        ]
    },
    entry_points={
        "console_scripts":[
            "halfapi=halfapi.cli.cli:cli"
        ]
    },
    keywords="web-api development boilerplate",
    project_urls={
        "Source": "https://github.com/halfAPI/halfapi",
    }
)



""" HalfRoute

Child class of starlette.routing.Route
"""
from functools import partial, wraps

from typing import Callable, Coroutine, List, Dict
from types import FunctionType

from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.exceptions import HTTPException

from .logging import logger
from .lib.domain import MissingAclError, PathError, UnknownPathParameterType, \
    UndefinedRoute, UndefinedFunction

class HalfRoute(Route):
    """ HalfRoute
    """
    def __init__(self, path: List[str], fct: Callable, params: List[Dict], method: str):
        logger.info('HalfRoute creation: %s %s %s %s', path, fct, params, method)
        if len(params) == 0:
            raise MissingAclError('[{}] {}'.format(method, '/'.join(path)))

        if len(path) == 0:
            logger.error('Empty path for [{%s}]', method)
            raise PathError()

        super().__init__(
            path,
            HalfRoute.acl_decorator(
                fct,
                params
            ),
            methods=[method])

    @staticmethod
    def acl_decorator(fct: Callable = None, params: List[Dict] = None) -> Coroutine:
        """
        Decorator for async functions that calls pre-conditions functions
        and appends kwargs to the target function


        Parameters:
            fct (Callable):
                The function to decorate

            params List[Dict]:
                A list of dicts that have an "acl" key that points to a function

        Returns:
            async function
        """

        if not params:
            params = []

        if not fct:
            return partial(HalfRoute.acl_decorator, params=params)


        @wraps(fct)
        async def caller(req: Request, *args, **kwargs):
            for param in params:
                if param.get('acl'):
                    passed = param['acl'](req, *args, **kwargs)
                    if isinstance(passed, FunctionType):
                        passed = param['acl']()(req, *args, **kwargs)

                    if not passed:
                        logger.debug(
                            'ACL FAIL for current route (%s - %s)', fct, param.get('acl'))
                        continue

                    logger.debug(
                        'ACL OK for current route (%s - %s)', fct, param.get('acl'))

                    req.scope['acl_pass'] = param['acl'].__name__

                    if 'args' in param:
                        req.scope['args'] = param['args']
                        logger.debug(
                            'Args for current route (%s)', param.get('args'))

                    if 'out' in param:
                        req.scope['out'] = param['out']

                    if 'out' in param:
                        req.scope['out'] = param['out'].copy()

                    if 'check' in req.query_params:
                        return PlainTextResponse(param['acl'].__name__)

                    logger.debug('acl_decorator %s', param)
                    logger.debug('calling %s:%s %s %s', fct.__module__, fct.__name__, args, kwargs)
                    return await fct(
                        req, *args,
                        **{
                            **kwargs,
                        })

            if 'check' in req.query_params:
                return PlainTextResponse('')

            raise HTTPException(401)

        return caller



# from starlette.routing import Route
# from halfapi.lib.routes import gen_starlette_routes, gen_router_routes
# 
# def test_gen_starlette_routes():
#     from .dummy_domain import routers
#     for route in gen_starlette_routes({
#         'dummy_domain': routers }):
# 
#         assert isinstance(route, Route)
# 
# import pytest
# 
# @pytest.mark.skip
# def test_api_routes():
#     from . import dummy_domain
#     d_res, d_acls = api_routes(dummy_domain)
#     assert isinstance(d_res, dict)
#     assert isinstance(d_acls, dict)
# 
#     yielded = False
# 
#     for path, verb, m_router, fct, params in gen_router_routes(dummy_domain, []):
#         if not yielded:
#             yielded = True
# 
#         assert path in d_res
#         assert verb in d_res[path]
#         assert 'docs' in d_res[path][verb]
#         assert 'acls' in d_res[path][verb]
#         assert isinstance(d_res[path][verb]['docs'], dict)
#         assert isinstance(d_res[path][verb]['acls'], list)
#         assert len(d_res[path][verb]['acls']) == len(params)
# 
#     assert yielded is True



from unittest import TestCase
import sys
import pytest
from halfapi.halfapi import HalfAPI

class TestConf(TestCase):
    def setUp(self):
        self.args = {
            'domain': {
                'dummy_domain': {
                    'name': 'dummy_domain',
                    'router': '.routers',
                    'enabled': True,
                    'prefix': False,
                }
            }
        }
    def tearDown(self):
        pass

    def test_conf_production_default(self):
        halfapi = HalfAPI({
            **self.args
        })
        assert halfapi.PRODUCTION is True

    def test_conf_production_true(self):
        halfapi = HalfAPI({
            **self.args,
            'production': True,
        })
        assert halfapi.PRODUCTION is True

    def test_conf_production_false(self):
        halfapi = HalfAPI({
            **self.args,
            'production': False,
        })
        assert halfapi.PRODUCTION is False

    def test_conf_variables(self):
        from halfapi.conf import (
            CONFIG,
            SCHEMA,
        )

        assert isinstance(CONFIG, dict)
        assert isinstance(CONFIG.get('project_name'), str)
        assert isinstance(SCHEMA, dict)
        assert isinstance(CONFIG.get('secret'), str)
        assert isinstance(CONFIG.get('host'), str)
        assert isinstance(CONFIG.get('port'), int)



#!/usr/bin/env python3
"""
Routes module

Classes :
    - JSONRoute

Fonctions :
    - gen_domain_routes
    - gen_schema_routes
    - api_routes

Exception :
    - DomainNotFoundError

"""
import inspect

from typing import Coroutine, Dict, Generator, Tuple, Any
from types import ModuleType, FunctionType

import yaml

# from .domain import gen_router_routes, domain_acls, route_decorator, domain_schema
from .responses import ORJSONResponse
from .acl import args_check
from ..half_route import HalfRoute
from . import acl

from ..logging import logger

class DomainNotFoundError(Exception):
    """ Exception when a domain is not importable
    """

def JSONRoute(data: Any) -> Coroutine:
    """
    Returns a route function that returns the data as JSON

    Parameters:
        data (Any):
            The data to return

    Returns:
        async function
    """
    async def wrapped(request, *args, **kwargs):
        return ORJSONResponse(data)

    return wrapped


def gen_domain_routes(m_domain: ModuleType):
    """
    Yields the Route objects for a domain

    Parameters:
        m_domains: ModuleType

    Returns:
        Generator(HalfRoute)
    """
    yield HalfRoute('/',
        JSONRoute(domain_schema(m_domain)),
        [{'acl': acl.public}],
        'GET'
    )

    for path, method, m_router, fct, params in gen_router_routes(m_domain, []):
        yield HalfRoute(f'/{path}', fct, params, method)


def gen_schema_routes(schema: Dict):
    """
    Yields the Route objects according to a given schema
    """
    for path, methods in schema.items():
        for verb, definition in methods.items():
            fct = definition.pop('fct')
            acls = definition.pop('acls')
            # TODO: Check what to do with gen_routes, it is almost the same function
            if not inspect.iscoroutinefunction(fct):
                yield HalfRoute(path, route_decorator(fct), acls, verb)
            else:
                yield HalfRoute(path, args_check(fct), acls, verb)


def api_routes(m_dom: ModuleType) -> Tuple[Dict, Dict]:
    """
    Yields the description objects for HalfAPI app routes

    Parameters:
        m_dom (ModuleType): the halfapi module

    Returns:
        (Dict, Dict)
    """

    d_acls = {}

    def str_acl(params):
        l_params = []

        for param in params:

            if 'acl' not in param.keys() or not param['acl']:
                continue

            l_params.append(param.copy())
            l_params[-1]['acl'] = param['acl'].__name__

            if param['acl'] not in d_acls.keys():
                d_acls[param['acl'].__name__] = param['acl']

        return l_params

    d_res = {}
    for path, verb, m_router, fct, params in gen_router_routes(m_dom, []):
        try:
            if path not in d_res:
                d_res[path] = {}

            d_res[path][verb] = {
                'docs': yaml.load(fct.__doc__, Loader=yaml.FullLoader),
                'acls': str_acl(params)
            }
        except Exception as exc:
            logger.error("""Error in route generation
                path:%s
                verb:%s
                router:%s
                fct:%s
                params:%s """, path, verb, m_router, fct, params)
            raise exc

    return d_res, d_acls
